#include "nvfs.h"

static void get_xattr_range(struct nvfs_memory_inode *nmi, struct nvfs_xattr_entry **start, struct nvfs_xattr_entry **end)
{
	struct nvfs_inode *nvi = nmi->nvi;
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	uint64_t block = le64_to_cpu(nvi->i_xattr_block);
	if (block == 0) {
		*start = nvi->i_xattr;
		*end = (void *)nvi + (1U << nvs->log2_inode_size);
	} else {
		*start = nvs->memory_map + block;
		*end = nvs->memory_map + block + (1U << nvs->log2_page_size);
	}
}

static struct nvfs_xattr_entry *find_xattr_entry(struct nvfs_memory_inode *nmi, unsigned prefix, const char *name, size_t name_len)
{
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	struct nvfs_xattr_entry *start, *end;

	get_xattr_range(nmi, &start, &end);

	do {
		if (nvfs_xattr_get_prefix(start) == prefix &&
		    nvfs_xattr_get_name_len(start) == name_len &&
		    !memcmp(start->data, name, name_len)) {
			return start;
		}
		start = (void *)start + nvfs_xattr_entry_size(start);
	} while (start < end);
	if (unlikely(start > end)) {
		nvfs_error(nvs, "corrupted extended attributes");
		return ERR_PTR(-EUCLEAN);
	}
	return NULL;
}

static struct nvfs_xattr_entry *nvfs_xattr_prepare_page(struct nvfs_memory_inode *nmi, const char *name, size_t name_len, const void *value, size_t value_len)
{
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	struct nvfs_xattr_entry *start, *end;
	size_t needed_len = nvfs_xattr_entry_needed_size(name_len, value_len);

	get_xattr_range(nmi, &start, &end);

	do {
		size_t de_size = nvfs_xattr_entry_size(start);
		if (nvfs_xattr_get_prefix(start) != NVFS_XATTR_PREFIX_EMPTY)
			goto used;
		if (de_size < needed_len) {
			struct nvfs_xattr_entry *next = (void *)start + de_size;
			size_t next_size;
			if (next >= end)
				goto used;
			if (nvfs_xattr_get_prefix(next) != NVFS_XATTR_PREFIX_EMPTY)
				goto used;
			next_size = de_size + nvfs_xattr_entry_size(next);
			nv_store(start->x, nvfs_xattr_entry_make_size(next_size));
			wmb();
			continue;
		}
		if (de_size > needed_len) {
			struct nvfs_xattr_entry *next = (void *)start + needed_len;
			nv_store(next->x, nvfs_xattr_entry_make_size(de_size - needed_len));
			wmb();
		}

		nv_store(start->x, nvfs_xattr_entry_make_size(needed_len));
		memcpy_flushcache(start->data, name, name_len);
		memcpy_flushcache(start->data + name_len, value, value_len);

		return start;

used:
		start = (void *)start + nvfs_xattr_entry_size(start);
	} while (start < end);

	if (unlikely(start > end)) {
		nvfs_error(nvs, "corrupted extended attributes");
		return ERR_PTR(-EUCLEAN);
	}

	return NULL;
}

static int nvfs_defragment_entries(struct nvfs_superblock *nvs, struct nvfs_xattr_entry *dst, struct nvfs_xattr_entry *src)
{
	struct nvfs_xattr_entry *dst_end = (void *)dst + (1U << nvs->log2_page_size);
	struct nvfs_xattr_entry *src_end = (void *)src + (1U << nvs->log2_page_size);

	while (src < src_end) {
		unsigned pfx = nvfs_xattr_get_prefix(src);
		unsigned size = nvfs_xattr_entry_size(src);
		if (pfx == NVFS_XATTR_PREFIX_EMPTY)
			goto next;

		memcpy_flushcache(dst, src, size);
		dst += size;

next:
		src = (struct nvfs_xattr_entry *)((char *)src + size);
	}

	if (unlikely(src > src_end)) {
		nvfs_error(nvs, "corrupted extended attributes");
		return -EUCLEAN;
	}

	if (dst < dst_end) {
		nv_store(dst->x, nvfs_xattr_entry_make_size(dst_end - dst));
	}

	return 0;
}

static struct nvfs_xattr_entry *nvfs_xattr_prepare_add(struct nvfs_memory_inode *nmi, const char *name, size_t name_len, const void *value, size_t value_len)
{
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	struct nvfs_xattr_entry *nxe;
	bool defragmented = false;
	uint64_t ob;

again:
	nxe = nvfs_xattr_prepare_page(nmi, name, name_len, value, value_len);
	if (nxe)
		return nxe;

	ob = le64_to_cpu(nmi->nvi->i_xattr_block);
	if (!ob) {
		struct nvfs_xattr_entry *start, *end, *end2;
		uint64_t b;

		b = nvfs_new_block(nvs, NULL);
		if (!b)
			return ERR_PTR(-ENOSPC);

		get_xattr_range(nmi, &start, &end);
		memcpy_flushcache(nvs->memory_map + b, start, (char *)end - (char *)start);
		end2 = nvs->memory_map + b + ((char *)end - (char *)start);
		nv_store(end2->x, nvfs_xattr_entry_make_size((1U << nvs->log2_page_size) - ((char *)end - (char *)start)));
		wmb();
		nv_store(nmi->nvi->i_xattr_block, cpu_to_le64(b));
		wmb();
		goto again;
	}

	if (!defragmented) {
		struct nvfs_xattr_entry *start, *end;
		uint64_t b;
		int r;

		b = nvfs_new_block(nvs, NULL);
		if (!b)
			return ERR_PTR(-ENOSPC);

		get_xattr_range(nmi, &start, &end);

		r = nvfs_defragment_entries(nvs, nvs->memory_map + b, start);
		if (unlikely(r)) {
			nvfs_free_blocks(nvs, b, 1);
			return ERR_PTR(r);
		}

		wmb();
		nv_store(nmi->nvi->i_xattr_block, cpu_to_le64(b));
		wmb();
		nvfs_free_blocks(nvs, ob, 1);

		defragmented = true;
		goto again;
	}

	return ERR_PTR(-ENOSPC);
}

static void nvfs_xattr_finish_add(struct nvfs_xattr_entry *nxe, unsigned prefix, size_t name_len, size_t value_len)
{
	uint32_t val;

	wmb();

	val =	((name_len + value_len) << NVFS_XATTR_TOTAL_LEN_OFFSET) |
		(prefix << (NVFS_XATTR_PREFIX_OFFSET - NVFS_XATTR_PREFIX_SHIFT)) |
		(name_len << NVFS_XATTR_NAMELEN_OFFSET);

	nv_store(nxe->x, cpu_to_le32(val));
}


static void nvfs_xattr_delete(struct nvfs_xattr_entry *nxe)
{
	uint32_t val;
	val = le32_to_cpu(nxe->x);
	val &= NVFS_XATTR_TOTAL_LEN_MASK;
	nv_store(nxe->x, cpu_to_le32(val));
}


static bool nvfs_xattr_trusted_list(struct dentry *dentry)
{
	return capable(CAP_SYS_ADMIN);
}

int nvfs_xattr_get(struct inode *inode, int type, const char *name,
		   void *buffer, size_t buffer_len)
{
	struct nvfs_memory_inode *nmi = i_to_nmi(inode);
	struct nvfs_xattr_entry *nxe;
	int r;
	size_t data_len;
	size_t name_len = strlen(name);

	down_read(&nmi->xattr_sem);

	if (unlikely(name_len > NVFS_XATTR_NAMELEN_MASK >> NVFS_XATTR_NAMELEN_OFFSET)) {
		r = -ERANGE;
		goto ret;
	}

	nxe = find_xattr_entry(i_to_nmi(inode), type, name, name_len);
	if (!nxe) {
		r = -ENODATA;
		goto ret;
	}
	if (unlikely(IS_ERR(nxe))) {
		r = PTR_ERR(nxe);
		goto ret;
	}

	data_len = nvfs_xattr_get_total_len(nxe) - name_len;
	if (buffer) {
		if (unlikely(data_len > buffer_len)) {
			r = -ERANGE;
			goto ret;
		}
		memcpy(buffer, nxe->data + name_len, data_len);
	}
	r = data_len;

ret:
	up_read(&nmi->xattr_sem);
	return r;
}

static int nvfs_xattr_get_handler(const struct xattr_handler *handler,
				  struct dentry *unused, struct inode *inode,
				  const char *name, void *buffer, size_t buffer_len)
{
	return nvfs_xattr_get(inode, handler->flags, name, buffer, buffer_len);
}

int nvfs_xattr_set(struct inode *inode, int type, const char *name,
		   const void *buffer, size_t buffer_len, int flags)
{
	struct nvfs_memory_inode *nmi = i_to_nmi(inode);
	struct nvfs_superblock *nvs = inode->i_sb->s_fs_info;
	struct nvfs_xattr_entry *nxe, *new_nxe = NULL /* avoid warning */;
	int r;
	size_t name_len = strlen(name);

	if (buffer == NULL)
		buffer_len = 0;

	down_write(&nmi->xattr_sem);

	if (unlikely(name_len > NVFS_XATTR_NAMELEN_MASK >> NVFS_XATTR_NAMELEN_OFFSET) ||
	    unlikely(buffer_len > 1U << nvs->log2_page_size)) {
		r = -ERANGE;
		goto ret;
	}

	nxe = find_xattr_entry(i_to_nmi(inode), type, name, name_len);
	if (IS_ERR(nxe)) {
		r = PTR_ERR(nxe);
		goto ret;
	}
	if (nxe) {
		if (flags & XATTR_CREATE) {
			r = -EEXIST;
			goto ret;
		}
	} else {
		if (flags & XATTR_REPLACE) {
			r = -ENODATA;
			goto ret;
		}
	}

	if (buffer) {
		new_nxe = nvfs_xattr_prepare_add(i_to_nmi(inode), name, name_len, buffer, buffer_len);
		if (unlikely(IS_ERR(new_nxe))) {
			r = PTR_ERR(new_nxe);
			goto ret;
		}
	}

	if (nxe) {
		/*
		 * we must search for it again, because nvfs_xattr_prepare_add may have
		 * moved it to a page
		 */
		nxe = find_xattr_entry(i_to_nmi(inode), type, name, name_len);
		if (unlikely(IS_ERR(nxe))) {
			r = PTR_ERR(nxe);
			goto ret;
		}
		if (unlikely(!nxe)) {
			nvfs_error(nvs, "xattr entry disappeared");
			r = -EUCLEAN;
			goto ret;
		}
		nvfs_xattr_delete(nxe);
	}

	if (buffer) {
		nvfs_xattr_finish_add(new_nxe, type, name_len, buffer_len);
	}

	r = 0;

ret:
	up_write(&nmi->xattr_sem);
	return r;
}

static int nvfs_xattr_set_handler(const struct xattr_handler *handler,
				  struct dentry *unused, struct inode *inode,
				  const char *name, const void *buffer,
				  size_t buffer_len, int flags)
{
	return nvfs_xattr_set(inode, handler->flags, name, buffer, buffer_len, flags);
}

ssize_t nvfs_listxattr(struct dentry *dentry, char *list, size_t list_size)
{
	struct nvfs_memory_inode *nmi = i_to_nmi(d_inode(dentry));
	struct nvfs_superblock *nvs = nmi->vfs_inode.i_sb->s_fs_info;
	struct nvfs_xattr_entry *start, *end;
	ssize_t r = 0;

	down_read(&nmi->xattr_sem);

	get_xattr_range(nmi, &start, &end);

	do {
		const char *prefix_name;
		size_t prefix_len, name_len, len;
		const struct xattr_handler **handler_p;
		unsigned pfx = nvfs_xattr_get_prefix(start);
		if (pfx == NVFS_XATTR_PREFIX_EMPTY)
			goto next;

		for (handler_p = nvfs_xattr_handlers; *handler_p; handler_p++) {
			if ((*handler_p)->flags == pfx)
				goto found_handler;
		}
		goto next;

found_handler:
		if ((*handler_p)->list && !(*handler_p)->list(dentry))
			goto next;
		prefix_name = (*handler_p)->prefix ?: (*handler_p)->name;
		prefix_len = strlen(prefix_name);
		name_len = nvfs_xattr_get_name_len(start);
		len = prefix_len + name_len + 1;
		if (!list) {
			r += len;
		} else {
			if (unlikely(list_size < len)) {
				r = -ERANGE;
				goto ret;
			}
			memcpy(list, prefix_name, prefix_len);
			memcpy(list + prefix_len, start->data, name_len);
			list[prefix_len + name_len] = 0;
			list += len;
			list_size -= len;
			r += len;
		}

next:
		start = (void *)start + nvfs_xattr_entry_size(start);
	} while (start < end);
	if (unlikely(start > end)) {
		nvfs_error(nvs, "corrupted extended attributes");
		r = -EUCLEAN;
		goto ret;
	}

ret:
	up_read(&nmi->xattr_sem);
	return r;
}


static int nvfs_initxattrs(struct inode *inode, const struct xattr *xattr_array,
			   void *fs_info)
{
	const struct xattr *xattr;
	int err = 0;

	for (xattr = xattr_array; xattr->name != NULL; xattr++) {
		err = nvfs_xattr_set(inode, NVFS_XATTR_PREFIX_SECURITY,
				     xattr->name, xattr->value,
				     xattr->value_len, 0);
		if (err < 0)
			break;
	}
	return err;
}

int nvfs_init_security(struct inode *inode, struct inode *dir,
		       const struct qstr *qstr)
{
	return security_inode_init_security(inode, dir, qstr,
					    &nvfs_initxattrs, NULL);
}


static const struct xattr_handler nvfs_xattr_user_handler = {
	.prefix = XATTR_USER_PREFIX,
	.get	= nvfs_xattr_get_handler,
	.set	= nvfs_xattr_set_handler,
	.flags	= NVFS_XATTR_PREFIX_USER,
};

static const struct xattr_handler nvfs_xattr_trusted_handler = {
	.prefix = XATTR_TRUSTED_PREFIX,
	.list	= nvfs_xattr_trusted_list,
	.get	= nvfs_xattr_get_handler,
	.set	= nvfs_xattr_set_handler,
	.flags	= NVFS_XATTR_PREFIX_TRUSTED,
};

static const struct xattr_handler nvfs_xattr_security_handler = {
	.prefix = XATTR_SECURITY_PREFIX,
	.get	= nvfs_xattr_get_handler,
	.set	= nvfs_xattr_set_handler,
	.flags	= NVFS_XATTR_PREFIX_SECURITY,
};

const struct xattr_handler *nvfs_xattr_handlers[] = {
	&nvfs_xattr_user_handler,
	&nvfs_xattr_trusted_handler,
#ifdef CONFIG_NVFS_POSIX_ACL
	&posix_acl_access_xattr_handler,
	&posix_acl_default_xattr_handler,
#endif
	&nvfs_xattr_security_handler,
	NULL
};
